Attribute VB_Name = "Triangulation"
Option Explicit

' Make a BSP tree holding the triangles.
Public Function MakeBspTree(ByVal num_triangles As Integer, triangles() As SimplePolygon) As BSPTree
Dim bsp_tree As BSPTree
Dim i As Integer
Dim num_triangles1 As Integer
Dim num_triangles2 As Integer
Dim triangles1() As SimplePolygon
Dim triangles2() As SimplePolygon

    ' Make the tree's root, splitting with
    ' the first triangle.
    Set bsp_tree = New BSPTree
    Set bsp_tree.SplittingTriangle = triangles(1)
    bsp_tree.SplittingTriangle.Finish

    ' See how many triangles there are.
    If num_triangles > 1 Then
        ' Split the remaining triangles.
        For i = 2 To num_triangles
            ' Add this triangle to the tree.
            bsp_tree.AddTriangle triangles(i)
        Next i
    End If
End Function
' Triangulate a polygon.
Public Sub TriangulatePolygon(ByRef num_triangles As Integer, triangles() As SimplePolygon, ByVal pgon As SimplePolygon)
Dim pt As Integer

    ' Orient the polygon.
    pgon.OrientPolygonClockwise

    ' While the polygon has more than
    ' three points, remove an ear.
    Do While pgon.PointX.Count > 3
        ' Remove an ear from the polygon.
        RemoveEar num_triangles, triangles, pgon
    Loop

    ' The polygon is now a triangle, too.
    num_triangles = num_triangles + 1
    ReDim Preserve triangles(1 To num_triangles)
    Set triangles(num_triangles) = pgon
End Sub
' Remove an ear from the polygon.
Private Sub RemoveEar(ByRef num_triangles As Integer, triangles() As SimplePolygon, ByVal pgon As SimplePolygon)
Dim A As Integer
Dim B As Integer
Dim C As Integer
Dim tmp As Integer

    ' Find an ear.
    FindEar A, B, C, pgon

    ' Create a new triangle for the ear.
    num_triangles = num_triangles + 1
    ReDim Preserve triangles(1 To num_triangles)
    Set triangles(num_triangles) = New SimplePolygon
    With triangles(num_triangles)
        .AddPoint pgon.PointX(A), pgon.PointY(A), pgon.PointZ(A)
        .AddPoint pgon.PointX(B), pgon.PointY(B), pgon.PointZ(B)
        .AddPoint pgon.PointX(C), pgon.PointY(C), pgon.PointZ(C)

        .FillColor = pgon.FillColor
        .ForeColor = pgon.ForeColor
    End With

    ' Remove the ear from the polygon
    ' by removing point B.
    pgon.PointX.Remove B
    pgon.PointY.Remove B
    pgon.PointZ.Remove B
End Sub

' Return True if the three points form an ear.
Private Function FormsEar(ByVal A As Integer, ByVal B As Integer, ByVal C As Integer, ByVal pgon As SimplePolygon) As Boolean
Dim num_pts As Integer
Dim pt As Integer
Dim triangle As SimplePolygon

    ' Assume the points form an ear.
    FormsEar = True

    ' See if the angle ABC is concave.
    If GetAngle( _
        pgon.PointX(A), pgon.PointY(A), _
        pgon.PointX(B), pgon.PointY(B), _
        pgon.PointX(C), pgon.PointY(C)) > 0 _
    Then
        ' This is a concave corner so the triangle
        ' cannot be an ear.
        FormsEar = False
        Exit Function
    End If

    ' Make the triangle A, B, C.
    Set triangle = New SimplePolygon
    triangle.AddPoint pgon.PointX(A), pgon.PointY(A), pgon.PointZ(A)
    triangle.AddPoint pgon.PointX(B), pgon.PointY(B), pgon.PointZ(B)
    triangle.AddPoint pgon.PointX(C), pgon.PointY(C), pgon.PointZ(C)

    ' Check the other points to see if they lie in the
    ' triangle A, B, C.
    num_pts = pgon.PointX.Count
    For pt = 1 To num_pts
        If pt <> A And pt <> B And pt <> C Then
            If PointInPolygon(pgon.PointX(pt), pgon.PointY(pt), triangle) Then
                ' This point is in the triangle so
                ' this is not an ear.
                FormsEar = False
                Exit For
            End If
        End If
    Next pt
End Function
' Return True if the point is in the polygon.
Public Function PointInPolygon(ByVal X As Single, ByVal Y As Single, ByVal pgon As SimplePolygon) As Boolean
Dim num_pts As Integer
Dim pt As Integer
Dim total_angle As Single

    ' Get the angle between the point and the
    ' first and last vertices.
    num_pts = pgon.PointX.Count
    total_angle = GetAngle( _
        pgon.PointX(num_pts), pgon.PointY(num_pts), _
        X, Y, _
        pgon.PointX(1), pgon.PointY(1))

    ' Add the angles from the point to each other
    ' pair of vertices.
    For pt = 1 To num_pts - 1
        total_angle = total_angle + GetAngle( _
            pgon.PointX(pt), pgon.PointY(pt), _
            X, Y, _
            pgon.PointX(pt + 1), pgon.PointY(pt + 1))
    Next pt

    ' The total angle should be 2 * PI or -2 * PI if
    ' the point is in the polygon and close to zero
    ' if the point is outside the polygon.
    PointInPolygon = (Abs(total_angle) > 0.000001)
End Function

' Return the angle with tangent opp/hyp. The returned
' value is between PI and -PI.
Public Function ATanB(ByVal opp As Single, ByVal adj As Single) As Single
Dim angle As Single

    ' Get the basic angle.
    If Abs(adj) < 0.0001 Then
        angle = PI / 2
    Else
        angle = Abs(Atn(opp / adj))
    End If

    ' See if we are in quadrant 2 or 3.
    If adj < 0 Then
        ' angle > PI/2 or angle < -PI/2.
        angle = PI - angle
    End If

    ' See if we are in quadrant 3 or 4.
    If opp < 0 Then
        angle = -angle
    End If

    ' Return the result.
    ATanB = angle
End Function

' Find the indexes of three points that form an ear.
Private Sub FindEar(ByRef A As Integer, ByRef B As Integer, ByRef C As Integer, ByVal pgon As SimplePolygon)
Dim num_pts As Integer

    num_pts = pgon.PointX.Count
    For A = 1 To num_pts
        B = A + 1
        If B > num_pts Then B = 1
        C = B + 1
        If C > num_pts Then C = 1
        If FormsEar(A, B, C, pgon) Then Exit Sub
    Next A

    ' We should never get here because there should
    ' always be at least two ears.
    Stop
End Sub

' Return the angle ABC.
' Return a value between PI and -PI.
' Note that the value is the opposite of what you might
' expect because Y coordinates increase downward.
Public Function GetAngle(ByVal Ax As Single, ByVal Ay As Single, ByVal Bx As Single, ByVal By As Single, ByVal Cx As Single, ByVal Cy As Single) As Single
Dim dot_product As Single
Dim cross_product As Single

    ' Get the dot product and cross product.
    dot_product = DotProduct(Ax, Ay, Bx, By, Cx, Cy)
    cross_product = CrossProductLength(Ax, Ay, Bx, By, Cx, Cy)

    ' Calculate the angle.
    GetAngle = ATanB(cross_product, dot_product)
End Function

' Return the cross product AB x BC.
' The cross product is a vector perpendicular to AB
' and BC having length |AB| * |BC| * Sin(theta) and
' with direction given by the right-hand rule.
' For two vectors in the X-Y plane, the result is a
' vector with X and Y components 0 so the Z component
' gives the vector's length and direction.
Public Function CrossProductLength( _
    ByVal Ax As Single, ByVal Ay As Single, _
    ByVal Bx As Single, ByVal By As Single, _
    ByVal Cx As Single, ByVal Cy As Single _
  ) As Single
Dim BAx As Single
Dim BAy As Single
Dim BCx As Single
Dim BCy As Single

    ' Get the vectors' coordinates.
    BAx = Ax - Bx
    BAy = Ay - By
    BCx = Cx - Bx
    BCy = Cy - By

    ' Calculate the Z coordinate of the cross product.
    CrossProductLength = BAx * BCy - BAy * BCx
End Function
' Return the dot product AB  BC.
' Note that AB  BC = |AB| * |BC| * Cos(theta).
Private Function DotProduct( _
    ByVal Ax As Single, ByVal Ay As Single, _
    ByVal Bx As Single, ByVal By As Single, _
    ByVal Cx As Single, ByVal Cy As Single _
  ) As Single
Dim BAx As Single
Dim BAy As Single
Dim BCx As Single
Dim BCy As Single

    ' Get the vectors' coordinates.
    BAx = Ax - Bx
    BAy = Ay - By
    BCx = Cx - Bx
    BCy = Cy - By

    ' Calculate the dot product.
    DotProduct = BAx * BCx + BAy * BCy
End Function


